# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreos@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import re
import numpy as np

from xmipp_metadata.metadata import XmippMetaData
from xmipp_metadata.image_handler import ImageHandler

import pyworkflow.protocol.params as params
from pyworkflow.object import String, Integer, CsvList, Boolean
from pyworkflow.utils.path import moveFile
from pyworkflow import VERSION_2_0

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
import pwem.emlib.metadata as md
from pwem.constants import ALIGN_PROJ
from pwem.objects import Volume, ParticleFlex, SetOfParticlesFlex

from xmipp3.convert import createItemMatrix, setXmippAttributes, writeSetOfParticles, \
    geometryFromMatrix, matrixFromGeometry
import xmipp3

import flexutils
import flexutils.constants as const
from flexutils.utils import getXmippFileName


class TensorflowProtAngularAlignmentHetSiren(ProtAnalysis3D, ProtFlexBase):
    """ Protocol for angular alignment with heterogeneous reconstruction with the HetSIREN algorithm."""
    _label = 'flexible align - HetSIREN'
    _lastUpdateVersion = VERSION_2_0

    # --------------------------- DEFINE param functions -----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addHidden(params.USE_GPU, params.BooleanParam, default=True,
                       label="Use GPU for execution",
                       help="This protocol has both CPU and GPU implementation.\
                             Select the one you want to use.")
        form.addHidden(params.GPU_LIST, params.StringParam, default='0',
                       expertLevel=params.LEVEL_ADVANCED,
                       label="Choose GPU IDs",
                       help="Add a list of GPU devices that can be used")
        group = form.addGroup("Data")
        group.addParam('inputParticles', params.PointerParam, label="Input particles", pointerClass='SetOfParticles')
        group.addParam('inputVolume', params.PointerParam, allowsNull=True,
                       label="Starting volume", pointerClass='Volume',
                       help="If provided, the HomoSIREN network will learn to refine with the new learned angles. "
                            "Otherwise, the network will learn the reconstruction of the map from scratch")
        group.addParam('inputVolumeMask', params.PointerParam,
                       label="Reconstruction mask", pointerClass='VolumeMask', allowsNull=True,
                       help="If provided, the pose refinement and reconstruction learned by HomoSIREN will be focused "
                            "in the region delimited by the mask. Otherise, a sphere inscribed in the volume box will "
                            "be used")
        group.addParam('boxSize', params.IntParam, default=128,
                       label='Downsample particles to this box size', expertLevel=params.LEVEL_ADVANCED,
                       help='In general, downsampling the particles will increase performance without compromising '
                            'the estimation the deformation field for each particle. Note that output particles will '
                            'have the original box size, and Zernike3D coefficients will be modified to work with the '
                            'original size images')
        group.addParam('outSize', params.IntParam, allowsNull=True,
                       label='Decoded volume size', expertLevel=params.LEVEL_ADVANCED,
                       help='Determines the box size of the volumes to be decoded by the network (i.e. the maximum '
                            'resolution achievable). If empty, it will match the downsampled box size. Otherwise, '
                            'it must be set to a value higher than or equal to the downsampled box size')
        group.addParam('trainSize', params.IntParam, allowsNull=True,
                       label='Image training size', expertLevel=params.LEVEL_ADVANCED,
                       help='By default, the size of the images used to train the network will match the value '
                            'specified for the downsampling parameter. However, in many cases it is useful to perform '
                            'a multi-resolution training by presenting the network first a further downsampled version '
                            'of the images to posteriorly perform a fine tuning on higher resolutions. This parameter '
                            'controls the current resolution/box size the network will see during training. If empty, '
                            'training image size will match the downsampling size. Otherwise, it must be set to a '
                            'number smaller than or equal to the downsampled size')
        group = form.addGroup("Latent Space")
        group.addParam('hetDim', params.IntParam, default=10, label='Latent space dimension',
                       expertLevel=params.LEVEL_ADVANCED,
                       help="Dimension of the HetSIREN bottleneck (latent space dimension)")
        group = form.addGroup("CTF Parameters (Advanced)",
                              expertLevel=params.LEVEL_ADVANCED)
        group.addParam('applyCTF', params.BooleanParam, default=True, label='Apply CTF?',
                       expertLevel=params.LEVEL_ADVANCED,
                       help="If true, volume projection will be subjected to CTF corrections")
        group.addParam('ctfType', params.EnumParam, choices=['Apply', 'Wiener'],
                       default=0, label="CTF correction type",
                       display=params.EnumParam.DISPLAY_HLIST,
                       condition="applyCTF", expertLevel=params.LEVEL_ADVANCED,
                       help="* *Apply*: CTF is applied to the projection generated from the reference map\n"
                            "* *Wiener*: input particle is CTF corrected by a Wiener fiter")
        group.addParam("pad", params.IntParam, default=2,
                       label="Padding factor",
                       condition="applyCTF", expertLevel=params.LEVEL_ADVANCED,
                       help="Determines the padding factor to be applied before computing "
                            "the Fourier Transform of the images to increase the frequency "
                            "content")
        form.addSection(label='Network')
        form.addParam('fineTune', params.BooleanParam, default=False, label="Fine tune previous network?",
                      help="If True, a previously trained deepPose network will be fine tuned based on the "
                           "new input parameters. Note that when this option is set, the input particles "
                           "must have a trained deepPose network associated (i.e. particles must come from "
                           "a **'angular align - deepPose'** protocol.")
        form.addParam('netProtocol', params.PointerParam, label="Previously trained network",
                      allowsNull=True,
                      pointerClass='TensorflowProtAngularAlignmentHetSiren',
                      condition="fineTune")
        group = form.addGroup("Network hyperparameters")
        group.addParam('architecture', params.EnumParam, choices=['DeepConv', 'ConvNN', 'MPLNN'],
                       expertLevel=params.LEVEL_ADVANCED,
                       default=1, label="Network architecture", display=params.EnumParam.DISPLAY_HLIST,
                       help="* *DeepConv*: a deep convolution neural architecture based on ResNet principles\n"
                            "* *ConvNN*: convolutional neural network\n"
                            "* *MLPNN*: multiperceptron neural network")
        group.addParam('stopType', params.EnumParam, choices=['Samples', 'Manual'],
                       default=1, label="How to compute total epochs?",
                       display=params.EnumParam.DISPLAY_HLIST,
                       help="* *Samples*: Epochs will be obtained from the total number of samples "
                            "the network will see\n"
                            "* *Epochs*: Total number of epochs is provided manually")
        group.addParam('epochs', params.IntParam, default=20, condition="stopType==1",
                       label='Number of training epochs')
        group.addParam('maxSamples', params.IntParam, default=1000000, condition="stopType==0",
                       label="Samples",
                       help='Maximum number of samples seen during network training')
        group.addParam('batch_size', params.IntParam, default=8, label='Number of images in batch',
                       help="Number of images that will be used simultaneously for every training step. "
                            "We do not recommend to change this value unless you experience memory errors. "
                            "In this case, value should be decreased.")
        group.addParam('lr', params.FloatParam, default=1e-5, label='Learning rate',
                       help="The learning rate determines how fast the network will train based on the "
                            "seen samples. The larger the value, the faster the network although divergence "
                            "might occur. We recommend decreasing the learning rate value if this happens.")
        group.addParam('xla', params.BooleanParam, default=True, label="Allow XLA compilation?",
                       help="When XLA compilation is allowed, extra optimizations are applied during neural network "
                            "training increasing the training performance. However, XLA will only work with compatible "
                            "GPUs. If any error is experienced, set to No.")
        group.addParam('tensorboard', params.BooleanParam, default=True, label="Allow Tensorboard visualization?",
                       help="Tensorboard visualization provides a complete real-time report to supervides the training "
                            "of the neural network. However, for very large networks RAM requirements to save the "
                            "Tensorboard logs might overflow. If your process unexpectedly finishes when saving the "
                            "network callbacks, please, set this option to NO and restart the training.")
        group = form.addGroup("Extra network parameters")
        group.addParam('refinePose', params.BooleanParam, default=True, label="Refine pose?",
                       help="If True, the neural network will be also trained to refine the angular "
                            "and shift assignation of the particles to make it more consistent with the "
                            "heterogeneity estimation. Otherwise, only heterogeneity information will be "
                            "estimated.")
        group.addParam('split_train', params.FloatParam, default=1.0, label='Traning dataset fraction',
                       help="This value (between 0 and 1) determines the fraction of images that will "
                            "be used to train the network.")
        group.addParam('step', params.IntParam, default=1, label='Points step',
                       help="How many points (voxels) to skip during the training computations. "
                            "A value of 1 means that all point within the mask provided in the input "
                            "will be used. A value of 2 implies that half of the point will be skipped "
                            "to increase the performance.")
        group = form.addGroup("Logger")
        group.addParam('debugMode', params.BooleanParam, default=False, label='Debugging mode',
                       help="If you experience any error during the training execution, we recommend setting "
                            "this parameter to True followed by a restart of this protocol to generate a more "
                            "informative logging file.")
        form.addSection(label='Cost function')
        form.addParam('costFunction', params.EnumParam, choices=['MSE', 'Correlation', 'FPC'],
                      default=0, label="Cost function type", display=params.EnumParam.DISPLAY_HLIST,
                      help="Determine the cost function to be minimized during the neural network training. Both, "
                           "**Correlation** and **FPC** will yield similar results. However, **Corr * FPC** "
                           "allows excluding high frequency information by masking in the Fourier space. This might "
                           "help preveting overfitting in scenarios with low Signal to Noise ratios at the expense "
                           "of slightly increasing computation time.\n"
                           "Recommended cost function is **MSE** unlsess a reference map has been specified. "
                           "If a volume is going to be refined, **Correlation** or **FPC** will provide more accurate "
                           "results")
        form.addParam('maskRadius', params.FloatParam, default=0.85, label="Mask radius (%)",
                      condition="costFunction==2",
                      help="Determine the radius (in percentage) of the circular mask to be applied to the Fourier "
                           "Transform of the images. A value of 1 implies that the circular mask is inscribed to the "
                           "bounding box the Fourier Transform.")
        form.addParam("smoothMask", params.BooleanParam, default=True, label="Smooth mask?",
                      condition="costFunction==2",
                      help="If True, the mask applied to the Fourier Transform of the particle images will have a smooth"
                           "vanishing transition.")
        form.addParam("l1Reg", params.FloatParam, default=0.1, label="L1 loss regularization",
                      help="Determines the weight of the L1 map minimization in the cost function. L1 is moslty used to "
                           "decrease the amount of noise in the map learned by the network. We do not recommend to touch "
                           "this parameter")
        form.addParam("tvReg", params.FloatParam, default=0.1, label="Total variation loss regularization",
                      help="Determines the weight of the TV map minimization in the cost function. TV is moslty used to "
                           "promote densitiy smoothness while focusing on the preservation of edges present in "
                           "the CryoEM map.")
        form.addParam("mseReg", params.FloatParam, default=0.1, label="MSE loss regularization",
                      help="Determines the weight of the MSE variation map minimization in the cost function. "
                           "MSE is moslty used to promote density smoothnes while focusing on ensuring a continuous "
                           "transition of the density values")
        form.addParam("multires", params.FloatParam, allowsNull=True, label="Multiresolution levels",
                      help="Determines the number of multiresolution filter used to compare the reconstructed map and "
                           "the experimental images at different resolutions. If empty, no multiresolution strategy is "
                           "applied in the cost function. Multiresolution helps making the training more robust by "
                           "sacrifying the resolution of the decoded maps.")
        form.addSection(label='Output')
        form.addParam("filterDecoded", params.BooleanParam, default=True, label="Filter decoded map?",
                      help="If True, the maps decoded after training the network will be convoluted with a Gaussian filter. "
                           "In general, this postprocessing is not needed unless 'Points step' parameter is set to a value "
                           "greater than 1")
        form.addParam("onlyPos", params.BooleanParam, default=False, label="Remove negative values?",
                      help="If True, the negative values from map decoded after training the network will be removed.")
        form.addParam("numVol", params.IntParam, default=20, label="Number of decoded maps",
                      help="Determines in how many regions the trained latent space will be splitted by "
                           "KMeans, allowing to decode a state based on the representative of each cluster. "
                           "This provides and initial summary/exploration of the trained landscape")
        form.addParallelSection(threads=4, mpi=0)

    def _createFilenameTemplates(self):
        """ Centralize how files are called """
        myDict = {
            'imgsFn': self._getExtraPath('input_particles.xmd'),
            'fnVol': self._getExtraPath('volume.mrc'),
            'fnVolMask': self._getExtraPath('mask.mrc'),
            'fnOutDir': self._getExtraPath()
        }
        self._updateFilenamesDict(myDict)

    # --------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._createFilenameTemplates()
        self._insertFunctionStep(self.writeMetaDataStep)
        self._insertFunctionStep(self.trainingStep)
        self._insertFunctionStep(self.predictStep)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions -----------------------
    def writeMetaDataStep(self):
        imgsFn = self._getFileName('imgsFn')
        fnVol = self._getFileName('fnVol')
        fnVolMask = self._getFileName('fnVolMask')
        md_file = self._getFileName('imgsFn')

        inputParticles = self.inputParticles.get()
        Xdim = inputParticles.getXDim()
        self.newXdim = self.boxSize.get()
        self.vol_mask_dim = self.outSize.get() if self.outSize.get() is not None else self.newXdim

        if self.inputVolume.get():  # Map reference
            ih = ImageHandler()
            inputVolume = self.inputVolume.get().getFileName()
            ih.convert(getXmippFileName(inputVolume), fnVol)
            curr_vol_dim = ImageHandler(getXmippFileName(inputVolume)).getDimensions()[-1]
            if curr_vol_dim != self.vol_mask_dim:
                self.runJob("xmipp_image_resize",
                            "-i %s --dim %d " % (fnVol, self.vol_mask_dim), numberOfMpi=1, env=xmipp3.Plugin.getEnviron())

        if self.inputVolumeMask.get():  # Mask reference
            ih = ImageHandler()
            inputMask = self.inputVolumeMask.get().getFileName()
            if inputMask:
                ih.convert(getXmippFileName(inputMask), fnVolMask)
                curr_mask_dim = ImageHandler(getXmippFileName(inputMask)).getDimensions()[-1]
                if curr_mask_dim != self.vol_mask_dim:
                    self.runJob("xmipp_image_resize",
                                "-i %s --dim %d --interp nearest" % (fnVolMask, self.vol_mask_dim), numberOfMpi=1,
                                env=xmipp3.Plugin.getEnviron())
        else:
            ImageHandler().createCircularMask(fnVolMask, boxSize=self.vol_mask_dim, is3D=True)

        writeSetOfParticles(inputParticles, imgsFn)

        # Write extra attributes (if needed)
        md = XmippMetaData(md_file)
        if hasattr(inputParticles.getFirstItem(), "_xmipp_subtomo_labels"):
            labels = np.asarray([int(particle._xmipp_subtomo_labels) for particle in inputParticles.iterItems()])
            md[:, "subtomo_labels"] = labels
        md.write(md_file, overwrite=True)

        if self.newXdim != Xdim:
            params = "-i %s -o %s --save_metadata_stack %s --fourier %d" % \
                     (imgsFn,
                      self._getTmpPath('scaled_particles.stk'),
                      self._getExtraPath('scaled_particles.xmd'),
                      self.newXdim)
            if self.numberOfMpi.get() > 1:
                params += " --mpi_job_size %d" % int(inputParticles.getSize() / self.numberOfMpi.get())
            self.runJob("xmipp_image_resize", params, numberOfMpi=self.numberOfMpi.get(),
                        env=xmipp3.Plugin.getEnviron())
            moveFile(self._getExtraPath('scaled_particles.xmd'), imgsFn)

    def trainingStep(self):
        md_file = self._getFileName('imgsFn')
        out_path = self._getExtraPath('network')
        if not os.path.isdir(out_path):
            os.mkdir(out_path)
        inputParticles = self.inputParticles.get()
        pad = self.pad.get()
        batch_size = self.batch_size.get()
        step = self.step.get()
        split_train = self.split_train.get()
        lr = self.lr.get()
        l1Reg = self.l1Reg.get()
        tvReg = self.tvReg.get()
        mseReg = self.mseReg.get()
        hetDim = self.hetDim.get()
        self.newXdim = self.boxSize.get()
        correctionFactor = self.inputParticles.get().getXDim() / self.newXdim
        sr = correctionFactor * self.inputParticles.get().getSamplingRate()
        trainSize = self.trainSize.get() if self.trainSize.get() is not None else self.newXdim
        if isinstance(inputParticles, SetOfParticlesFlex) and hasattr(inputParticles.getFlexInfo(), "outSize"):
            outSize = inputParticles.getFlexInfo().outSize.get()
        else:
            outSize = self.outSize.get() if self.outSize.get() is not None else self.newXdim
        applyCTF = self.applyCTF.get()
        xla = self.xla.get()
        tensorboard = self.tensorboard.get()
        args = "--md_file %s --out_path %s --batch_size %d " \
               "--shuffle --split_train %f --pad %d " \
               "--sr %f --apply_ctf %d --step %d --l1_reg %f --tv_reg %f --mse_reg %f --het_dim %d --lr %f " \
               "--trainSize %d --outSize %d" \
               % (md_file, out_path, batch_size, split_train, pad, sr, applyCTF, step,
                  l1Reg, tvReg, mseReg, hetDim, lr, trainSize, outSize)

        if self.stopType.get() == 0:
            args += " --max_samples_seen %d" % self.maxSamples.get()
        else:
            args += " --epochs %d" % self.epochs.get()

        if self.costFunction.get() == 0:
            args += " --cost mse"
        elif self.costFunction.get() == 1:
            args += " --cost corr"
        elif self.costFunction.get() == 2:
            args += " --cost corr-fpc --radius_mask %f" % self.maskRadius.get()
            if self.smoothMask.get():
                args += " --smooth_mask"

        if self.architecture.get() == 0:
            args += " --architecture deepconv"
        elif self.architecture.get() == 1:
            args += " --architecture convnn"
        elif self.architecture.get() == 2:
            args += " --architecture mlpnn"

        if self.ctfType.get() == 0:
            args += " --ctf_type apply"
        elif self.ctfType.get() == 1:
            args += " --ctf_type wiener"

        if self.refinePose.get():
            args += " --refine_pose"

        if self.onlyPos.get():
            args += " --only_pos"

        if self.multires.get():
            args += " --multires %d" % self.multires.get()

        if self.fineTune.get():
            netProtocol = self.netProtocol.get()
            modelPath = netProtocol._getExtraPath(os.path.join('network', 'het_siren_model.h5'))
            args += " --weigths_file %s" % modelPath

        if xla:
            args += " --jit_compile"

        if tensorboard:
            args += " --tensorboard"

        if self.useGpu.get():
            gpu_list = ','.join([str(elem) for elem in self.getGpuList()])
            args += " --gpu %s" % gpu_list

        if self.debugMode.get():
            log_level = 0
        else:
            log_level = 2

        program = flexutils.Plugin.getTensorflowProgram("train_het_siren.py", python=False,
                                                        log_level=log_level)
        self.runJob(program, args, numberOfMpi=1)

    def predictStep(self):
        md_file = self._getFileName('imgsFn')
        weigths_file = self._getExtraPath(os.path.join('network', 'het_siren_model.h5'))
        inputParticles = self.inputParticles.get()
        pad = self.pad.get()
        hetDim = self.hetDim.get()
        numVol = self.numVol.get()
        self.newXdim = self.boxSize.get()
        correctionFactor = self.inputParticles.get().getXDim() / self.newXdim
        sr = correctionFactor * self.inputParticles.get().getSamplingRate()
        trainSize = self.trainSize.get() if self.trainSize.get() is not None else self.newXdim
        if isinstance(inputParticles, SetOfParticlesFlex) and hasattr(inputParticles.getFlexInfo(), "outSize"):
            outSize = inputParticles.getFlexInfo().outSize.get()
        else:
            outSize = self.outSize.get() if self.outSize.get() is not None else self.newXdim
        applyCTF = self.applyCTF.get()
        args = "--md_file %s --weigths_file %s --pad %d --sr %f " \
               "--apply_ctf %d --het_dim %d --num_vol %d --trainSize %d --outSize %d" \
               % (md_file, weigths_file, pad, sr, applyCTF, hetDim, numVol, trainSize, outSize)

        if self.ctfType.get() == 0:
            args += " --ctf_type apply"
        elif self.ctfType.get() == 1:
            args += " --ctf_type wiener"

        if self.architecture.get() == 0:
            args += " --architecture deepconv"
        elif self.architecture.get() == 1:
            args += " --architecture convnn"
        elif self.architecture.get() == 2:
            args += " --architecture mlpnn"

        if self.refinePose.get():
            args += " --refine_pose"

        if self.filterDecoded.get():
            args += " --apply_filter"

        if self.onlyPos.get():
            args += " --only_pos"

        if self.useGpu.get():
            gpu_list = ','.join([str(elem) for elem in self.getGpuList()])
            args += " --gpu %s" % gpu_list

        program = flexutils.Plugin.getTensorflowProgram("predict_het_siren.py", python=False)
        self.runJob(program, args, numberOfMpi=1)

    def createOutputStep(self):
        inputParticles = self.inputParticles.get()
        Xdim = inputParticles.getXDim()
        self.newXdim = self.boxSize.get()
        trainSize = self.trainSize.get() if self.trainSize.get() is not None else self.newXdim
        if isinstance(inputParticles, SetOfParticlesFlex) and hasattr(inputParticles.getFlexInfo(), "outSize"):
            outSize = inputParticles.getFlexInfo().outSize.get()
        else:
            outSize = self.outSize.get() if self.outSize.get() is not None else self.newXdim
        model_path = self._getExtraPath(os.path.join('network', 'het_siren_model.h5'))
        md_file = self._getFileName('imgsFn')

        metadata = XmippMetaData(md_file)
        latent_space = np.asarray([np.fromstring(item, sep=',') for item in metadata[:, 'latent_space']])
        delta_rot = metadata[:, 'delta_angle_rot']
        delta_tilt = metadata[:, 'delta_angle_tilt']
        delta_psi = metadata[:, 'delta_angle_psi']
        delta_shift_x = metadata[:, 'delta_shift_x']
        delta_shift_y = metadata[:, 'delta_shift_y']

        inputSet = self.inputParticles.get()
        partSet = self._createSetOfParticlesFlex(progName=const.HETSIREN)

        partSet.copyInfo(inputSet)
        partSet.setHasCTF(inputSet.hasCTF())
        partSet.setAlignmentProj()

        correctionFactor = Xdim / self.newXdim

        inverseTransform = partSet.getAlignment() == ALIGN_PROJ

        idx = 0
        for particle in inputSet.iterItems():
            outParticle = ParticleFlex(progName=const.HETSIREN)
            outParticle.copyInfo(particle)

            outParticle.setZFlex(latent_space[idx])

            tr_ori = particle.getTransform().getMatrix()
            shifts, angles = geometryFromMatrix(tr_ori, inverseTransform)

            # Apply delta angles
            angles[0] += delta_rot[idx]
            angles[1] += delta_tilt[idx]
            angles[2] += delta_psi[idx]

            # Apply delta shifts
            shifts[0] += correctionFactor * delta_shift_x[idx]
            shifts[1] += correctionFactor * delta_shift_y[idx]

            # Set new transformation matrix
            tr = matrixFromGeometry(shifts, angles, inverseTransform)
            outParticle.getTransform().setMatrix(tr)

            partSet.append(outParticle)

            idx += 1

        partSet.getFlexInfo().modelPath = String(model_path)
        partSet.getFlexInfo().coordStep = Integer(self.step.get())
        partSet.getFlexInfo().outSize = Integer(outSize)
        partSet.getFlexInfo().trainSize = Integer(trainSize)

        if self.inputVolume.get():
            inputVolume = self.inputVolume.get().getFileName()
            partSet.getFlexInfo().refMap = String(inputVolume)

        if self.inputVolumeMask.get():
            inputMask = self.inputVolumeMask.get().getFileName()
            partSet.refMask = String(inputMask)

        if self.architecture.get() == 0:
            partSet.getFlexInfo().architecture = String("deepconv")
        elif self.architecture.get() == 1:
            partSet.getFlexInfo().architecture = String("convnn")
        elif self.architecture.get() == 2:
            partSet.getFlexInfo().architecture = String("mlpnn")

        if self.ctfType.get() == 0:
            partSet.getFlexInfo().ctfType = String("apply")
        elif self.ctfType.get() == 1:
            partSet.getFlexInfo().ctfType = String("wiener")

        partSet.getFlexInfo().pad = Integer(self.pad.get())

        outVols = self._createSetOfVolumes()
        outVols.setSamplingRate(inputParticles.getSamplingRate())
        for idx in range(self.numVol.get()):
            outVol = Volume()
            outVol.setSamplingRate(inputParticles.getSamplingRate())

            ImageHandler().scaleSplines(self._getExtraPath('decoded_map_class_%02d.mrc' % (idx + 1)),
                                        self._getExtraPath('decoded_map_class_%02d.mrc' % (idx + 1)),
                                        finalDimension=inputParticles.getXDim(), overwrite=True)

            outVol.setLocation(self._getExtraPath('decoded_map_class_%02d.mrc' % (idx + 1)))
            outVols.append(outVol)

        self._defineOutputs(outputParticles=partSet)
        self._defineTransformRelation(self.inputParticles, partSet)

        self._defineOutputs(outputVolumes=outVols)
        self._defineTransformRelation(self.inputParticles, outVols)

    # --------------------------- UTILS functions -----------------------
    def _updateParticle(self, item, row):
        setXmippAttributes(item, row, md.MDL_ANGLE_ROT, md.MDL_ANGLE_TILT,
                           md.MDL_ANGLE_PSI, md.MDL_SHIFT_X, md.MDL_SHIFT_Y,
                           md.MDL_FLIP)
        createItemMatrix(item, row, align=ALIGN_PROJ)

    def getInputParticles(self):
        return self.inputParticles.get()

    def readPDB(self, fnIn):
        with open(fnIn) as f:
            lines = f.readlines()
        return lines

    def PDB2List(self, lines):
        newlines = []
        for line in lines:
            eval = re.search(r'^ATOM\s+\d+\s+/N|CA|C|O/\s+', line) if self.onlyBackbone.get() else line.startswith(
                "ATOM ")
            if eval:
                try:
                    x = float(line[30:38])
                    y = float(line[38:46])
                    z = float(line[46:54])
                    val = float(line[54:60])
                    newline = [x, y, z, val]
                    newlines.append(newline)
                except:
                    pass
        return newlines

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        logFile = os.path.abspath(self._getLogsPath()) + "/run.stdout"
        with open(logFile, "r") as fi:
            for ln in fi:
                if ln.startswith("GPU memory has"):
                    summary.append(ln)
                    break
        return summary

    # ----------------------- VALIDATE functions -----------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []

        inputParticles = self.inputParticles.get()
        boxSize = self.boxSize.get()
        trainSize = self.trainSize.get()
        if isinstance(inputParticles, SetOfParticlesFlex) and hasattr(inputParticles.getFlexInfo(), "outSize"):
            outSize = inputParticles.getFlexInfo().outSize
        else:
            outSize = self.outSize.get()

        if outSize is not None and outSize < boxSize:
            errors.append("Decoded image size must be larger than or equal to the downsampled box"
                          " size (currently set to %d)" % boxSize)

        if trainSize is not None and trainSize > boxSize:
            errors.append("Train image size must be smaller than or equal to the downsampled box"
                          " size (currently set to %d)" % boxSize)

        return errors



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreos@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import re
import numpy as np

from xmipp_metadata.metadata import XmippMetaData
from xmipp_metadata.image_handler import ImageHandler

import pyworkflow.protocol.params as params
from pyworkflow.object import String, Integer
from pyworkflow.utils.path import moveFile
from pyworkflow import VERSION_2_0
import pyworkflow.utils as pwutils

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
import pwem.emlib.metadata as md
from pwem.constants import ALIGN_PROJ
from pwem.objects import ParticleFlex

from xmipp3.convert import createItemMatrix, setXmippAttributes, writeSetOfParticles, \
    geometryFromMatrix, matrixFromGeometry
import xmipp3

import flexutils
import flexutils.constants as const
from flexutils.utils import getXmippFileName


class TensorflowProtDenoiseParticlesHetSiren(ProtAnalysis3D, ProtFlexBase):
    """ Denoise particles with the HetSIREN network. """
    _label = 'denoise particles - HetSIREN'
    _lastUpdateVersion = VERSION_2_0

    # --------------------------- DEFINE param functions -----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addHidden(params.USE_GPU, params.BooleanParam, default=True,
                       label="Use GPU for execution",
                       help="This protocol has both CPU and GPU implementation.\
                             Select the one you want to use.")
        form.addHidden(params.GPU_LIST, params.StringParam, default='0',
                       expertLevel=params.LEVEL_ADVANCED,
                       label="Choose GPU IDs",
                       help="Add a list of GPU devices that can be used")
        group = form.addGroup("Data")
        group.addParam('inputParticles', params.PointerParam, label="Input particles to predict",
                       pointerClass='SetOfParticles')
        group.addParam('hetSirenProtocol', params.PointerParam, label="HetSIREN trained network",
                       pointerClass='TensorflowProtAngularAlignmentHetSiren',
                       help="Previously executed 'angular align - HetSIREN'. "
                            "This will allow to load the network trained in that protocol to be used during "
                            "the prediction")
        form.addParallelSection(threads=4, mpi=0)

    def _createFilenameTemplates(self):
        """ Centralize how files are called """
        myDict = {
            'imgsFn': self._getExtraPath('input_particles.xmd'),
            'fnVol': self._getExtraPath('volume.mrc'),
            'fnVolMask': self._getExtraPath('mask.mrc'),
            'fnOutDir': self._getExtraPath()
        }
        self._updateFilenamesDict(myDict)

    # --------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._createFilenameTemplates()
        self._insertFunctionStep(self.writeMetaDataStep)
        self._insertFunctionStep(self.predictStep)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions -----------------------
    def writeMetaDataStep(self):
        imgsFn = self._getFileName('imgsFn')
        fnVol = self._getFileName('fnVol')
        fnVolMask = self._getFileName('fnVolMask')

        inputParticles = self.inputParticles.get()
        hetSirenProtocol = self.hetSirenProtocol.get()
        Xdim = inputParticles.getXDim()
        self.newXdim = hetSirenProtocol.boxSize.get()
        self.vol_mask_dim = hetSirenProtocol.outSize.get() if hetSirenProtocol.outSize.get() is not None else self.newXdim

        if hetSirenProtocol.inputVolume.get():  # Map reference
            ih = ImageHandler()
            inputVolume = hetSirenProtocol.inputVolume.get().getFileName()
            ih.convert(getXmippFileName(inputVolume), fnVol)
            curr_vol_dim = ImageHandler(getXmippFileName(inputVolume)).getDimensions()[-1]
            if curr_vol_dim != self.vol_mask_dim:
                self.runJob("xmipp_image_resize",
                            "-i %s --dim %d " % (fnVol, self.vol_mask_dim), numberOfMpi=1, env=xmipp3.Plugin.getEnviron())

        if hetSirenProtocol.inputVolumeMask.get():  # Mask reference
            ih = ImageHandler()
            inputMask = hetSirenProtocol.inputVolumeMask.get().getFileName()
            if inputMask:
                ih.convert(getXmippFileName(inputMask), fnVolMask)
                curr_mask_dim = ImageHandler(getXmippFileName(inputMask)).getDimensions()[-1]
                if curr_mask_dim != self.vol_mask_dim:
                    self.runJob("xmipp_image_resize",
                                "-i %s --dim %d --interp nearest" % (fnVolMask, self.vol_mask_dim), numberOfMpi=1,
                                env=xmipp3.Plugin.getEnviron())
        else:
            ImageHandler().createCircularMask(fnVolMask, boxSize=self.vol_mask_dim, is3D=True)

        writeSetOfParticles(inputParticles, imgsFn)

        if self.newXdim != Xdim:
            params = "-i %s -o %s --save_metadata_stack %s --fourier %d" % \
                     (imgsFn,
                      self._getTmpPath('scaled_particles.stk'),
                      self._getExtraPath('scaled_particles.xmd'),
                      self.newXdim)
            if self.numberOfMpi.get() > 1:
                params += " --mpi_job_size %d" % int(inputParticles.getSize() / self.numberOfMpi.get())
            self.runJob("xmipp_image_resize", params, numberOfMpi=self.numberOfMpi.get(),
                        env=xmipp3.Plugin.getEnviron())
            moveFile(self._getExtraPath('scaled_particles.xmd'), imgsFn)

    def predictStep(self):
        hetSirenProtocol = self.hetSirenProtocol.get()
        md_file = self._getFileName('imgsFn')
        weigths_file = hetSirenProtocol._getExtraPath(os.path.join('network', 'het_siren_model.h5'))
        pad = hetSirenProtocol.pad.get()
        self.newXdim = hetSirenProtocol.boxSize.get()
        Xdim = self.inputParticles.get().getXDim()
        correctionFactor = Xdim / self.newXdim
        sr = correctionFactor * self.inputParticles.get().getSamplingRate()
        applyCTF = hetSirenProtocol.ctfType.get()
        hetDim = hetSirenProtocol.hetDim.get()
        trainSize = hetSirenProtocol.trainSize.get() if hetSirenProtocol.trainSize.get() else self.newXdim
        outSize = hetSirenProtocol.outSize.get() if hetSirenProtocol.outSize.get() else self.newXdim
        args = "--md_file %s --weigths_file %s --pad %d " \
               "--sr %f --apply_ctf %d --het_dim %d --trainSize %d --outSize %d" \
               % (md_file, weigths_file, pad, sr, applyCTF, hetDim, trainSize, outSize)

        if hetSirenProtocol.ctfType.get() == 0:
            args += " --ctf_type apply"
        elif hetSirenProtocol.ctfType.get() == 1:
            args += " --ctf_type wiener"

        if hetSirenProtocol.architecture.get() == 0:
            args += " --architecture deepconv"
        elif hetSirenProtocol.architecture.get() == 1:
            args += " --architecture convnn"
        elif hetSirenProtocol.architecture.get() == 2:
            args += " --architecture mlpnn"

        if hetSirenProtocol.refinePose.get():
            args += " --refine_pose"

        if self.useGpu.get():
            gpu_list = ','.join([str(elem) for elem in self.getGpuList()])
            args += " --gpu %s" % gpu_list

        program = flexutils.Plugin.getTensorflowProgram("predict_particles_het_siren.py", python=False)
        self.runJob(program, args, numberOfMpi=1)

        # Scale particles
        ImageHandler().scaleSplines(self._getExtraPath('decoded_particles.mrcs'),
                                    self._getExtraPath('decoded_particles.mrcs'),
                                    finalDimension=Xdim, overwrite=True, isStack=True)

    def createOutputStep(self):
        inputParticles = self.inputParticles.get()
        hetSirenProtocol = self.hetSirenProtocol.get()
        Xdim = inputParticles.getXDim()
        self.newXdim = hetSirenProtocol.boxSize.get()
        trainSize = hetSirenProtocol.trainSize.get()
        outSize = hetSirenProtocol.outSize.get()
        model_path = hetSirenProtocol._getExtraPath(os.path.join('network', 'het_siren_model.h5'))
        md_file = self._getFileName('imgsFn')

        metadata = XmippMetaData(md_file)
        latent_space = np.asarray([np.fromstring(item, sep=',') for item in metadata[:, 'latent_space']])
        delta_rot = metadata[:, 'delta_angle_rot']
        delta_tilt = metadata[:, 'delta_angle_tilt']
        delta_psi = metadata[:, 'delta_angle_psi']
        delta_shift_x = metadata[:, 'delta_shift_x']
        delta_shift_y = metadata[:, 'delta_shift_y']
        denoised_images_path = metadata[:, "image"]

        inputSet = self.inputParticles.get()
        partSet = self._createSetOfParticlesFlex(progName=const.HETSIREN)

        partSet.copyInfo(inputSet)
        partSet.setAlignmentProj()

        correctionFactor = Xdim / self.newXdim

        inverseTransform = partSet.getAlignment() == ALIGN_PROJ

        idx = 0
        for particle in inputSet.iterItems():
            outParticle = ParticleFlex(progName=const.HETSIREN)
            outParticle.copyInfo(particle)

            index, _ = denoised_images_path[idx].split("@")
            outParticle.setLocation((int(index), self._getExtraPath("decoded_particles.mrcs")))

            outParticle.setZFlex(latent_space[idx])

            tr_ori = particle.getTransform().getMatrix()
            shifts, angles = geometryFromMatrix(tr_ori, inverseTransform)

            # Apply delta angles
            angles[0] += delta_rot[idx]
            angles[1] += delta_tilt[idx]
            angles[2] += delta_psi[idx]

            # Apply delta shifts
            shifts[0] += correctionFactor * delta_shift_x[idx]
            shifts[1] += correctionFactor * delta_shift_y[idx]

            # if refinePose:
            #     # Apply delta angles
            #     angles[0] += delta_rot[idx]
            #     angles[1] += delta_tilt[idx]
            #     angles[2] += delta_psi[idx]
            #
            #     # Apply delta shifts
            #     shifts[0] += correctionFactor * delta_shift_x[idx]
            #     shifts[1] += correctionFactor * delta_shift_y[idx]
            # else:
            #     # Assign delta angles
            #     angles[0] = delta_rot[idx]
            #     angles[1] = delta_tilt[idx]
            #     angles[2] = delta_psi[idx]
            #
            #     # Assign delta shifts
            #     shifts[0] = correctionFactor * delta_shift_x[idx]
            #     shifts[1] = correctionFactor * delta_shift_y[idx]

            # Set new transformation matrix
            tr = matrixFromGeometry(shifts, angles, inverseTransform)
            outParticle.getTransform().setMatrix(tr)

            partSet.append(outParticle)

            idx += 1

        partSet.getFlexInfo().modelPath = String(model_path)
        partSet.getFlexInfo().coordStep = Integer(hetSirenProtocol.step.get())
        partSet.getFlexInfo().outSize = Integer(outSize)
        partSet.getFlexInfo().trainSize = Integer(trainSize)

        if hetSirenProtocol.inputVolume.get():
            inputMask = hetSirenProtocol.inputVolumeMask.get().getFileName()
            inputVolume = hetSirenProtocol.inputVolume.get().getFileName()
            partSet.getFlexInfo().refMask = String(inputMask)
            partSet.getFlexInfo().refMap = String(inputVolume)

        if hetSirenProtocol.architecture.get() == 0:
            partSet.getFlexInfo().architecture = String("deepconv")
        elif hetSirenProtocol.architecture.get() == 1:
            partSet.getFlexInfo().architecture = String("convnn")
        elif hetSirenProtocol.architecture.get() == 2:
            partSet.getFlexInfo().architecture = String("mlpnn")

        if hetSirenProtocol.ctfType.get() == 0:
            partSet.getFlexInfo().ctfType = String("apply")
        elif hetSirenProtocol.ctfType.get() == 1:
            partSet.getFlexInfo().ctfType = String("wiener")

        partSet.getFlexInfo().pad = Integer(hetSirenProtocol.pad.get())

        self._defineOutputs(outputParticles=partSet)
        self._defineTransformRelation(self.inputParticles, partSet)

    # --------------------------- UTILS functions -----------------------
    def _updateParticle(self, item, row):
        setXmippAttributes(item, row, md.MDL_ANGLE_ROT, md.MDL_ANGLE_TILT,
                           md.MDL_ANGLE_PSI, md.MDL_SHIFT_X, md.MDL_SHIFT_Y,
                           md.MDL_FLIP)
        createItemMatrix(item, row, align=ALIGN_PROJ)

    def getInputParticles(self):
        return self.inputParticles.get()

    def readPDB(self, fnIn):
        with open(fnIn) as f:
            lines = f.readlines()
        return lines

    def PDB2List(self, lines):
        newlines = []
        for line in lines:
            eval = re.search(r'^ATOM\s+\d+\s+/N|CA|C|O/\s+', line) if self.onlyBackbone.get() else line.startswith(
                "ATOM ")
            if eval:
                try:
                    x = float(line[30:38])
                    y = float(line[38:46])
                    z = float(line[46:54])
                    val = float(line[54:60])
                    newline = [x, y, z, val]
                    newlines.append(newline)
                except:
                    pass
        return newlines

    # ----------------------- VALIDATE functions -----------------------
    def validate(self):
        """ Try to find errors on define params. """
        return []




# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreos@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import os
from xmipp_metadata.metadata import XmippMetaData
from xmipp_metadata.image_handler import ImageHandler

import pyworkflow.protocol.params as params
from pyworkflow.object import Integer, Float, String, Boolean
from pyworkflow.utils.path import moveFile
from pyworkflow import VERSION_2_0

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
import pwem.emlib.metadata as md
from pwem.constants import ALIGN_PROJ
from pwem.objects import ParticleFlex, SetOfParticlesFlex

from xmipp3.convert import (writeSetOfParticles, createItemMatrix,
                            setXmippAttributes, matrixFromGeometry)
import xmipp3

import flexutils.constants as const
import flexutils
from flexutils.utils import getXmippFileName


class XmippProtHeterogeneityPriorsZernike3D(ProtAnalysis3D, ProtFlexBase):
    """ Assignation of heterogeneity priors based on the Zernike3D basis. """
    _label = 'assign heterogeneity priors - Zernike3D'
    _lastUpdateVersion = VERSION_2_0

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        # form.addHidden(params.USE_GPU, params.BooleanParam, default=True,
        #                label="Use GPU for execution",
        #                help="This protocol has both CPU and GPU implementation.\
        #                    Select the one you want to use.")
        # form.addHidden(params.GPU_LIST, params.StringParam, default='0',
        #                expertLevel=params.LEVEL_ADVANCED,
        #                label="Choose GPU IDs",
        #                help="Add a list of GPU devices that can be used")
        form.addParam('inputParticles', params.PointerParam, label="Input particles", pointerClass='SetOfParticles')
        form.addParam('inputPriors', params.PointerParam, label="Input priors", pointerClass="SetOfVolumesFlex",
                      help="A set of volumes with Zernike3D coefficients associated to be used as priors for "
                           "the input particles")
        form.addParam('boxSize', params.IntParam, default=128,
                      label='Downsample particles to this box size', expertLevel=params.LEVEL_ADVANCED,
                      help='In general, downsampling the particles will increase performance without compromising '
                           'the estimation the deformation field for each particle. Note that output particles will '
                           'have the original box size, and Zernike3D coefficients will be modified to work with the '
                           'original size images')
        # form.addParam('maxShift', params.FloatParam, default=-1,
        #               label='Maximum shift (px)', expertLevel=params.LEVEL_ADVANCED,
        #               help='Maximum shift allowed in pixels')
        # form.addParam('maxAngular', params.FloatParam, default=5,
        #               label='Maximum angular change (degrees)', expertLevel=params.LEVEL_ADVANCED,
        #               help='Maximum angular change allowed (in degrees)')
        form.addParam('maxResolution', params.FloatParam,
                      label='Maximum resolution (A)', expertLevel=params.LEVEL_ADVANCED,
                      allowsNull=True,
                      help='Filter the particles to this sampling rate. By default, no filter is '
                           'applied')
        # form.addParam('regularization', params.FloatParam, default=0.005, label='Regularization',
        #               expertLevel=params.LEVEL_ADVANCED,
        #               help='Penalization to deformations (higher values penalize more the deformation).')
        form.addParam('ignoreCTF', params.BooleanParam, default=False, label='Ignore CTF?',
                      expertLevel=params.LEVEL_ADVANCED,
                      help="If true, volume projection won't be subjected to CTF corrections")
        # form.addParam('optimizeAlignment', params.BooleanParam, default=True, label='Optimize alignment?',
        #              expertLevel=params.LEVEL_ADVANCED)
        form.addParallelSection(threads=1, mpi=8)

    def _createFilenameTemplates(self):
        """ Centralize how files are called """
        myDict = {
            'imgsFn': self._getExtraPath('input_particles.xmd'),
            'fnVol': self._getExtraPath('input_volume.vol'),
            'fnVolMask': self._getExtraPath('input_volume_mask.vol'),
            'fnOut': self._getExtraPath('sphDone.xmd'),
            'fnPriors': self._getExtraPath("priors.txt"),
            'fnOutDir': self._getExtraPath()
                 }
        self._updateFilenamesDict(myDict)

    # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        self._createFilenameTemplates()
        self._insertFunctionStep("convertStep")
        self._insertFunctionStep("alignmentStep")
        self._insertFunctionStep("createOutputStep")

    # --------------------------- STEPS functions ---------------------------------------------------
    def convertStep(self):
        imgsFn = self._getFileName('imgsFn')
        fnVol = self._getFileName('fnVol')
        fnVolMask = self._getFileName('fnVolMask')

        inputParticles = self.inputParticles.get()
        inputPriors = self.inputPriors.get()
        writeSetOfParticles(inputParticles, imgsFn)
        Xdim = inputParticles.getXDim()
        # self.Ts = inputParticles.getSamplingRate()
        # newTs = self.targetResolution.get() * 1.0 / 3.0
        # self.newTs = max(self.Ts, newTs)
        # self.newXdim = int(Xdim * self.Ts / newTs)
        self.newXdim = self.boxSize.get()
        correctionFactor = self.newXdim / Xdim
        newTs = inputParticles.getSamplingRate() / correctionFactor
        np.savetxt(self._getExtraPath("sampling.txt"), [newTs])
        np.savetxt(self._getExtraPath("size.txt"), [self.newXdim])
        if self.newXdim != Xdim:
            self.runJob("xmipp_image_resize",
                        "-i %s -o %s --save_metadata_stack %s --fourier %d" %
                        (imgsFn,
                         self._getTmpPath('scaled_particles.stk'),
                         self._getExtraPath('scaled_particles.xmd'),
                         self.newXdim), numberOfMpi=1, env=xmipp3.Plugin.getEnviron())
            moveFile(self._getExtraPath('scaled_particles.xmd'), imgsFn)

        ih = ImageHandler()
        inputVolume = inputPriors.getFlexInfo().refMap.get()
        ih.convert(getXmippFileName(inputVolume), fnVol)
        # Xdim = self.inputParticles.get().getFirstItem().getDim()[0]
        if Xdim != self.newXdim:
            self.runJob("xmipp_image_resize",
                        "-i %s --dim %d " % (fnVol, self.newXdim), numberOfMpi=1,
                        env=xmipp3.Plugin.getEnviron())
        inputMask = inputPriors.getFlexInfo().refMask.get()
        if inputMask:
            ih.convert(getXmippFileName(inputMask), fnVolMask)
            if Xdim != self.newXdim:
                self.runJob("xmipp_image_resize",
                            "-i %s --dim %d --interp nearest" % (fnVolMask, self.newXdim), numberOfMpi=1,
                            env=xmipp3.Plugin.getEnviron())

    def alignmentStep(self):
        inputPriors = self.inputPriors.get()
        inputParticles =self.inputParticles.get()
        imgsFn = self._getFileName('imgsFn')
        fnVol = self._getFileName('fnVol')
        fnOut = self._getFileName('fnOut')
        fnVolMask = self._getFileName('fnVolMask')
        fnOutDir = self._getFileName('fnOutDir')
        fnPriors = self._getFileName('fnPriors')
        Ts = np.loadtxt(self._getExtraPath("sampling.txt"))
        maxResolution = self.maxResolution.get() if self.maxResolution.get() else Ts

        Xdim = inputParticles.getXDim()
        # self.Ts = inputParticles.getSamplingRate()
        # newTs = self.targetResolution.get() * 1.0 / 3.0
        # self.newTs = max(self.Ts, newTs)
        self.newXdim = self.boxSize.get()
        correctionFactor = self.newXdim / Xdim

        # Zernike3D parameters
        L1 = inputPriors.getFlexInfo().L1.get()
        L2 = inputPriors.getFlexInfo().L2.get()
        Rmax = correctionFactor * inputPriors.getFlexInfo().Rmax.get()

        # Write Zernike3D priors to file
        with open(fnPriors, 'w') as f:
            f.write(' '.join(map(str, [L1, L2, Rmax])) + "\n")
            for item in inputPriors.iterItems():
                z_clnm = item.getZFlex()
                z_clnm *= correctionFactor
                f.write(' '.join(map(str, z_clnm.reshape(-1))) + "\n")

        # Compute deformations
        def_file = self._getExtraPath("def_file.txt")
        args = "--i %s --z_clnm %s --o %s" % (fnVolMask, fnPriors, def_file)
        program = "compute_z_clnm_deformation.py"
        program = flexutils.Plugin.getProgram(program)
        self.runJob(program, args, numberOfMpi=1)
        deformations = np.loadtxt(def_file)

        # Write Zernike3D priors to file
        with open(fnPriors, 'w') as f:
            if deformations.size == 1:
                f.write(' '.join(map(str, [L1, L2, Rmax, deformations])) + "\n")
            else:
                f.write(' '.join(map(str, [L1, L2, Rmax] + deformations.tolist())) + "\n")
            for item in inputPriors.iterItems():
                z_clnm = item.getZFlex()
                z_clnm *= correctionFactor
                f.write(' '.join(map(str, z_clnm.reshape(-1))) + "\n")

        params = ' -i %s --ref %s -o %s ' \
                 '--l1 %d --l2 %d --sampling %f ' \
                 ' --max_resolution %f --odir %s --resume --regularization 0.0 --mask %s' \
                 ' --step 2 --blobr 2 --image_mode 1 --priors %s' %\
                 (imgsFn, fnVol, fnOut, L1, L2,
                  Ts, maxResolution, fnOutDir, fnVolMask, fnPriors)
        if not self.ignoreCTF.get():
            params += ' --useCTF'
        if self.inputParticles.get().isPhaseFlipped():
            params += ' --phaseFlipped'

        # if self.useGpu.get():
        #     params += ' --device %d' % self.getGpuList()[0]
        #     program = 'xmipp_cuda_angular_sph_alignment'
        #     self.runJob(program, params)
        # else:
        program = 'xmipp_forward_zernike_images_priors'
        self.runJob(program, params, numberOfMpi=self.numberOfMpi.get(), env=xmipp3.Plugin.getEnviron())


    def createOutputStep(self):
        inputParticles = self.getInputParticles()
        Xdim = inputParticles.getXDim()
        self.newXdim = self.boxSize.get()
        correctionFactor = Xdim / self.newXdim
        fnOut = self._getFileName('fnOut')
        mdOut = XmippMetaData(fnOut)

        # Zernike3D info
        inputPriors = self.inputPriors.get()
        L1 = inputPriors.getFlexInfo().L1.get()
        L2 = inputPriors.getFlexInfo().L2.get()
        Rmax = inputPriors.getFlexInfo().Rmax.get()
        inputVolume = inputPriors.getFlexInfo().refMap.get()
        inputMask = inputPriors.getFlexInfo().refMask.get()

        coeffs = correctionFactor * np.asarray([np.fromstring(item, sep=',') for item in mdOut[:, "sphCoefficients"]])
        deformation = correctionFactor * mdOut[:, "sphDeformation"]
        shifts = correctionFactor * mdOut[:, ["shiftX", "shiftY", "shiftZ"]]
        angles = mdOut[:, ["angleRot", "angleTilt", "anglePsi"]]

        partSet = self._createSetOfParticlesFlex(progName=const.ZERNIKE3D)

        partSet.copyInfo(inputParticles)
        partSet.setHasCTF(inputParticles.hasCTF())
        partSet.setAlignmentProj()

        inverseTransform = partSet.getAlignment() == ALIGN_PROJ

        idx = 0
        for particle in inputParticles.iterItems():
            outParticle = ParticleFlex(progName=const.ZERNIKE3D)
            outParticle.copyInfo(particle)

            outParticle.setZFlex(coeffs[idx])
            outParticle.getFlexInfo().deformation = Float(deformation[idx])

            # Set new transformation matrix
            tr = matrixFromGeometry(shifts[idx], angles[idx], inverseTransform)
            outParticle.getTransform().setMatrix(tr)

            partSet.append(outParticle)

            idx += 1

        partSet.getFlexInfo().L1 = Integer(L1)
        partSet.getFlexInfo().L2 = Integer(L2)
        partSet.getFlexInfo().Rmax = Float(Rmax)
        partSet.getFlexInfo().refMask = String(inputMask)
        partSet.getFlexInfo().refMap = String(inputVolume)

        self._defineOutputs(outputParticles=partSet)
        self._defineTransformRelation(self.inputParticles, partSet)


# --------------------------- UTILS functions --------------------------------------------
    def _updateParticle(self, item, row):
        setXmippAttributes(item, row, md.MDL_ANGLE_ROT, md.MDL_ANGLE_TILT,
                           md.MDL_ANGLE_PSI, md.MDL_SHIFT_X, md.MDL_SHIFT_Y,
                           md.MDL_FLIP, md.MDL_SPH_DEFORMATION,
                           md.MDL_SPH_COEFFICIENTS)
        createItemMatrix(item, row, align=ALIGN_PROJ)

    def getInputParticles(self):
        return self.inputParticles.get()

    # ----------------------- VALIDATE functions ----------------------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        inputPriors = self.inputPriors.get()
        if inputPriors.getFlexInfo().getProgName() != const.ZERNIKE3D:
            errors.append("The flexibility information associated with the priors is not "
                          "coming from the Zernike3D algorithm. Please, provide a set of priors "
                          "with the correct flexibility information.")
        return errors








# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import glob

import numpy as np
from pwem.protocols import ProtAnalysis3D
from pwem.objects import AtomStruct, SetOfAtomStructs, Volume

import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils
from pyworkflow.object import Float, Integer, String

import flexutils
from flexutils.utils import readZernikeFile
import flexutils.constants as const


class XmippMatchDeformSructZernike3D(ProtAnalysis3D):
    """ Zernike3D deformation field computation between two structures based on atomic matches"""
    _label = 'structure match and deform - Zernike3D'

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputs', params.MultiPointerParam, label="Input structures",
                      pointerClass='AtomStruct, SetOfAtomStructs',
                      help="Target structures")
        form.addParam('reference', params.PointerParam, label="Reference structure",
                      pointerClass='AtomStruct',
                      help="Structure to be deformed")
        form.addParam('l1', params.IntParam, default=3,
                      label='Zernike Degree',
                      expertLevel=params.LEVEL_ADVANCED,
                      help='Degree Zernike Polynomials of the deformation=1,2,3,...')
        form.addParam('l2', params.IntParam, default=2,
                      label='Harmonical Degree',
                      expertLevel=params.LEVEL_ADVANCED,
                      help='Degree Spherical Harmonics of the deformation=1,2,3,...')
        form.addParam('volume', params.PointerParam, label="Reference structure map",
                      pointerClass='Volume',
                      allowsNull=True,
                      expertLevel=params.LEVEL_ADVANCED,
                      help="Computed coefficients will be associated to this map instead to the "
                           "structure")
        form.addParam('mask', params.PointerParam, label="Reference map mask",
                      pointerClass='VolumeMask',
                      allowsNull=True,
                      expertLevel=params.LEVEL_ADVANCED,
                      help="Mask to associate to reference volume")
        form.addParam('moveBoxOrigin', params.BooleanParam, default=False, condition="volume",
                      label="Move structure to box origin?",
                      help="If PDB has been aligned inside Scipion, set to False. Otherwise, this option will "
                           "correctly place the PDB in the origin of the volume.")

    # --------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep("deformStep")
        self._insertFunctionStep("createOutputStep")

    # --------------------------- STEPS functions ------------------------------
    def deformStep(self):
        reference = self.reference.get().getFileName()
        l1 = self.l1.get()
        l2 = self.l2.get()
        if self.volume.get():
            volume = self.volume.get()
            rmax = int(0.5 * volume.getSamplingRate() * volume.getXDim())
        else:
            rmax = None

        input_files = []
        for pointer in self.inputs:
            obj = pointer.get()
            if isinstance(obj, AtomStruct):
                input_files.append(obj.getFileName())
            elif isinstance(obj, SetOfAtomStructs):
                for struct in obj:
                    input_files.append(struct.getFileName())

        for idp, input in enumerate(input_files):
            output = self._getExtraPath("structure_deformed_%d.pdb" % (idp + 1))
            args = "--i %s --r %s --o %s --l1 %d --l2 %d" \
                   % (reference, input, output, l1, l2)
            if rmax:
                args += " --rmax %d" % rmax
            if self.moveBoxOrigin.get() and rmax:
                args += " --d %d" % rmax
            elif not self.moveBoxOrigin.get() and rmax:
                args += " --d 0"
            program = "find_z_clnm_structure.py"
            program = flexutils.Plugin.getProgram(program)
            self.runJob(program, args)
            pwutils.moveFile(self._getExtraPath("z_clnm.txt"),
                             self._getExtraPath("z_clnm_%d.txt" % (idp + 1)))
            pwutils.moveFile(self._getExtraPath("rmsd_def.txt"),
                             self._getExtraPath("rmsd_def_%d.txt" % (idp + 1)))

    def createOutputStep(self):
        if self.volume.get():
            outputs = self._createSetOfVolumes()
            outputs.setSamplingRate(self.volume.get().getSamplingRate())
            Rmax = Float(int(0.5 * self.volume.get().getXDim()))
            outputs.refMap = String(self.volume.get().getFileName())
            outputs.refMask = String(self.mask.get().getFileName())
        else:
            outputs = SetOfAtomStructs().create(self._getPath())
            basis_params, _ = readZernikeFile(self._getExtraPath("z_clnm_1.txt"))
            Rmax = basis_params[2]

        L1 = Integer(self.l1.get())
        L2 = Integer(self.l2.get())

        outputs.L1 = L1
        outputs.L2 = L2
        outputs.Rmax = Rmax

        input_files = len(glob.glob(self._getExtraPath("z_clnm_*.txt")))
        for idp in range(input_files):
            z_clnm_file = self._getExtraPath("z_clnm_%d.txt" % (idp + 1))
            basis_params, z_clnm = readZernikeFile(z_clnm_file)
            rmsd_def = np.loadtxt(self._getExtraPath("rmsd_def_%d.txt" % (idp + 1)))
            deformation = Float(rmsd_def)

            print("Deformation for structure %d = %f" % (idp + 1, rmsd_def))

            if self.volume.get():
                output = Volume()
                output.setFileName(self.volume.get().getFileName())
                output.setSamplingRate(self.volume.get().getSamplingRate())
                z_clnm_vol = z_clnm[0] / output.getSamplingRate()
                output.L1 = L1
                output.L2 = L2
                output.Rmax = Rmax
                output._xmipp_sphDeformation = deformation
                output._xmipp_sphCoefficients = String(','.join(['%f' % c for c in z_clnm_vol]))
                output.refMap = String(output.getFileName())
                if self.mask.get():
                    output.refMask = String(self.mask.get().getFileName())
            else:
                outFile = self._getExtraPath("structure_deformed.pdb")
                output = AtomStruct(outFile)
                output.L1 = L1
                output.L2 = L2
                output.Rmax = Rmax
                output._xmipp_sphDeformation = deformation
                output._xmipp_sphCoefficients = String(','.join(['%f' % c for c in z_clnm[0]]))

            outputs.append(output)

        if self.volume.get():
            self._defineOutputs(zernikeVolumes=outputs)
        else:
            self._defineOutputs(zernikeStructures=outputs)
        self._defineSourceRelation(self.inputs, outputs)
        self._defineSourceRelation(self.reference, outputs)

    # --------------------------- UTILS functions ------------------------------



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreos@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import os

import pyworkflow.protocol.params as params
from pyworkflow.object import Boolean
from pyworkflow.utils.path import makePath
from pyworkflow import VERSION_2_0

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
from pwem.objects import ParticleFlex, SetOfParticlesFlex

import flexutils
from flexutils.utils import getOutputSuffix


class TensorflowProtInteractiveFlexConsensus(ProtAnalysis3D, ProtFlexBase):
    """ Protocol to filter particles based on a FlexConsensus network interactively """
    _label = 'interactive consensus - FlexConsensus'
    _lastUpdateVersion = VERSION_2_0
    OUTPUT_PREFIX = 'consensusParticles'

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addHidden(params.USE_GPU, params.BooleanParam, default=True,
                       label="Use GPU for execution",
                       help="This protocol has both CPU and GPU implementation.\
                             Select the one you want to use.")
        form.addHidden(params.GPU_LIST, params.StringParam, default='0',
                       expertLevel=params.LEVEL_ADVANCED,
                       label="Choose GPU IDs",
                       help="Add a list of GPU devices that can be used")
        group = form.addGroup("Data")
        group.addParam('inputSet', params.PointerParam,
                       label="Input particles", pointerClass='SetOfParticlesFlex')
        group.addParam('flexConsensusProtocol', params.PointerParam, label="FlexConsensus trained network",
                       pointerClass='TensorflowProtTrainFlexConsensus',
                       help="Previously executed 'train - FlexConsensus'. "
                            "This will allow to load the network trained in that protocol to be used during "
                            "the prediction")
        form.addParallelSection(threads=4, mpi=0)

    # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.convertInputStep)
        self._insertFunctionStep(self.predictStep)

    # --------------------------- STEPS functions ---------------------------------------------------
    def convertInputStep(self):
        particles = self.inputSet.get()
        data_path = self._getExtraPath("data")
        if not os.path.isdir(data_path):
            makePath(data_path)

        # Get data filename
        progName = particles.getFlexInfo().getProgName()
        data_file = progName + "_1.txt"

        # Read flexible space form particles
        z_flex = []
        for particle in particles.iterItems():
            z_flex.append(particle.getZFlex())
        z_flex = np.vstack(z_flex)

        # Save flexible space
        np.savetxt(os.path.join(data_path, data_file), z_flex)

    def predictStep(self):
        flexConsensusProtocol = self.flexConsensusProtocol.get()
        data_path = self._getExtraPath("data")
        out_path = self._getExtraPath()
        lat_dim = flexConsensusProtocol.latDim.get()
        weigths_file = flexConsensusProtocol._getExtraPath(os.path.join('network', 'flex_consensus_model.h5'))
        args = "--data_path %s --out_path %s --weigths_file %s --lat_dim %d" \
               % (data_path, out_path, weigths_file, lat_dim)

        if self.useGpu.get():
            gpu_list = ','.join([str(elem) for elem in self.getGpuList()])
            args += " --gpu %s" % gpu_list

        program = flexutils.Plugin.getTensorflowProgram("predict_flex_consensus.py", python=False)
        self.runJob(program, args, numberOfMpi=1)

    def _createOutput(self):
        inputSet = self.inputSet.get()
        selected_idx = np.loadtxt(self._getExtraPath("selected_idx.txt"))

        suffix = getOutputSuffix(self, SetOfParticlesFlex)
        partSet = self._createSetOfParticlesFlex(suffix, progName=inputSet.getFlexInfo().getProgName())

        partSet.copyInfo(inputSet)
        partSet.setHasCTF(inputSet.hasCTF())
        partSet.setAlignmentProj()

        idx = 0
        for particle in inputSet.iterItems():
            if idx in selected_idx:
                outParticle = ParticleFlex(progName=inputSet.getFlexInfo().getProgName())
                outParticle.copyInfo(particle)

                partSet.append(outParticle)

            idx += 1

        # Save new output
        name = self.OUTPUT_PREFIX + suffix
        args = {}
        args[name] = partSet
        self._defineOutputs(**args)
        self._defineSourceRelation(self.inputSet, partSet)

    # --------------------------- UTILS functions --------------------------------------------

    # ----------------------- VALIDATE functions ----------------------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []

        flexConsensusSets = self.flexConsensusProtocol.get().inputSets
        in_lat_dim = self.inputSet.get().getFirstItem().getZFlex().size
        dim_match = False

        for particle_set in flexConsensusSets:
            lat_dim = particle_set.get().getFirstItem().getZFlex().size
            if lat_dim == in_lat_dim:
                dim_match = True
                break

        if not dim_match:
            errors.append("The input particles' flexible information does not match the flexible space "
                          "dimension of the provided FlexConsensus network. Please, provide a set of particles "
                          "computed with some of the following programs:\n")
            progNames = []
            for particle_set in flexConsensusSets:
                progName = particle_set.get().getFlexInfo().getProgName()
                if progName not in progNames:
                    progNames.append(progName)
                    errors.append(f"     -{progName}")

        return errors



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreos@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import numpy as np
import re
from xmipp_metadata.metadata import XmippMetaData
import prody as pd

import pyworkflow.protocol.params as params
from pyworkflow.object import Integer, Float, String, CsvList, Boolean
from pyworkflow.utils.path import moveFile
from pyworkflow import VERSION_2_0

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
import pwem.emlib.metadata as md
from pwem.constants import ALIGN_PROJ
from pwem.objects import ParticleFlex

from xmipp3.convert import createItemMatrix, setXmippAttributes, writeSetOfParticles, \
                           geometryFromMatrix, matrixFromGeometry
import xmipp3

import flexutils
import flexutils.constants as const
from flexutils.protocols.xmipp.utils.custom_pdb_parser import PDBUtils


class TensorflowProtPredictDeepNMA(ProtAnalysis3D, ProtFlexBase):
    """ Predict NMA coefficents for a set of particles based on a trained
     DeepNMA network. """
    _label = 'predict - DeepNMA'
    _lastUpdateVersion = VERSION_2_0
    _subset = ["ca", "bb", "all"]

    # --------------------------- DEFINE param functions -----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addHidden(params.USE_GPU, params.BooleanParam, default=True,
                       label="Use GPU for execution",
                       help="This protocol has both CPU and GPU implementation.\
                             Select the one you want to use.")
        form.addHidden(params.GPU_LIST, params.StringParam, default='0',
                       expertLevel=params.LEVEL_ADVANCED,
                       label="Choose GPU IDs",
                       help="Add a list of GPU devices that can be used")
        group = form.addGroup("Data")
        group.addParam('inputParticles', params.PointerParam, label="Input particles to predict",
                      pointerClass='SetOfParticles')
        group.addParam('deepNMAProtocol', params.PointerParam, label="DeepNMA trained network",
                       pointerClass='TensorflowProtAngularAlignmentDeepNMA',
                       help="Previously executed 'angular align - DeepNMA'. "
                            "This will allow to load the network trained in that protocol to be used during "
                            "the prediction")
        form.addParallelSection(threads=4, mpi=0)

    def _createFilenameTemplates(self):
        """ Centralize how files are called """
        myDict = {
            'imgsFn': self._getExtraPath('input_particles.xmd'),
            'fnVol': self._getExtraPath('volume.mrc'),
            'fnVolMask': self._getExtraPath('mask.mrc'),
            'fnStruct': self._getExtraPath('structure.txt'),
            'fnConnect': self._getExtraPath("connectivity.txt"),
            'fnOutDir': self._getExtraPath()
        }
        self._updateFilenamesDict(myDict)

    # --------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._createFilenameTemplates()
        self._insertFunctionStep(self.writeMetaDataStep)
        self._insertFunctionStep(self.predictStep)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions -----------------------
    def writeMetaDataStep(self):
        imgsFn = self._getFileName('imgsFn')
        structure_file = self._getFileName('fnStruct')
        connect_file = self._getFileName('fnConnect')

        inputParticles = self.inputParticles.get()
        deepNMAProtocol = self.deepNMAProtocol.get()
        Xdim = inputParticles.getXDim()
        self.newXdim = deepNMAProtocol.boxSize.get()
        refStruct = deepNMAProtocol.inputStruct.get().getFileName()
        subset = self._subset[deepNMAProtocol.atomSubset.get()]

        # pd_struct = pd.parsePDB(refStruct, subset=subset, compressed=False)
        # pdb_coordinates = pd_struct.getCoords()
        # pdb_coordinates = np.c_[pdb_coordinates, np.ones(pdb_coordinates.shape[0])]
        # # pdb_lines = self.readPDB(deepNMAProtocol.inputStruct.get().getFileName())
        # # pdb_coordinates = np.array(self.PDB2List(pdb_lines))
        # np.savetxt(structure, pdb_coordinates)
        parser = PDBUtils(selectionString=subset)
        pdb_coordinates, connectivity = parser.parsePDB(refStruct)
        values = np.ones(pdb_coordinates.shape[0])
        pdb_coordinates = np.c_[pdb_coordinates, values]
        np.savetxt(structure_file, pdb_coordinates)
        np.savetxt(connect_file, connectivity)

        writeSetOfParticles(inputParticles, imgsFn)

        if self.newXdim != Xdim:
            params = "-i %s -o %s --save_metadata_stack %s --fourier %d" % \
                     (imgsFn,
                      self._getTmpPath('scaled_particles.stk'),
                      self._getExtraPath('scaled_particles.xmd'),
                      self.newXdim)
            if self.numberOfMpi.get() > 1:
                params += " --mpi_job_size %d" % int(inputParticles.getSize() / self.numberOfMpi.get())
            self.runJob("xmipp_image_resize", params, numberOfMpi=self.numberOfMpi.get(),
                        env=xmipp3.Plugin.getEnviron())
            moveFile(self._getExtraPath('scaled_particles.xmd'), imgsFn)

    def predictStep(self):
        deepNMAProtocol = self.deepNMAProtocol.get()
        md_file = self._getFileName('imgsFn')
        weigths_file = deepNMAProtocol._getExtraPath(os.path.join('network', 'deep_nma_model.h5'))
        n_modes = deepNMAProtocol.n_modes.get()
        pad = deepNMAProtocol.pad.get()
        correctionFactor = self.inputParticles.get().getXDim() / deepNMAProtocol.boxSize.get()
        sr = correctionFactor * self.inputParticles.get().getSamplingRate()
        applyCTF = deepNMAProtocol.applyCTF.get()
        args = "--md_file %s --weigths_file %s --n_modes %d " \
               "--pad %d --sr %f --apply_ctf %d" \
               % (md_file, weigths_file, n_modes, pad, sr, applyCTF)

        if deepNMAProtocol.refinePose.get():
            args += " --refine_pose"

        if deepNMAProtocol.ctfType.get() == 0:
            args += " --ctf_type apply"
        elif deepNMAProtocol.ctfType.get() == 1:
            args += " --ctf_type wiener"

        if deepNMAProtocol.architecture.get() == 0:
            args += " --architecture convnn"
        elif deepNMAProtocol.architecture.get() == 1:
            args += " --architecture mlpnn"


        if self.useGpu.get():
            gpu_list = ','.join([str(elem) for elem in self.getGpuList()])
            args += " --gpu %s" % gpu_list
        program = flexutils.Plugin.getTensorflowProgram("predict_deep_nma.py", python=False)
        self.runJob(program, args, numberOfMpi=1)

    def createOutputStep(self):
        inputParticles = self.inputParticles.get()
        deepNMAProtocol = self.deepNMAProtocol.get()
        Xdim = inputParticles.getXDim()
        self.newXdim = deepNMAProtocol.boxSize.get()
        model_path = deepNMAProtocol._getExtraPath(os.path.join('network', 'deep_nma_model.h5'))
        md_file = self._getFileName('imgsFn')

        metadata = XmippMetaData(md_file)
        nma_space = np.asarray([np.fromstring(item, sep=',') for item in metadata[:, 'nmaCoefficients']])

        if metadata.isMetaDataLabel('delta_angle_rot'):
            delta_rot = metadata[:, 'delta_angle_rot']
            delta_tilt = metadata[:, 'delta_angle_tilt']
            delta_psi = metadata[:, 'delta_angle_psi']
            delta_shift_x = metadata[:, 'delta_shift_x']
            delta_shift_y = metadata[:, 'delta_shift_y']

        refinePose = deepNMAProtocol.refinePose.get()

        inputSet = self.inputParticles.get()
        partSet = self._createSetOfParticlesFlex(progName=const.NMA)
        partSet.setHasCTF(inputSet.hasCTF())

        partSet.copyInfo(inputSet)
        partSet.setAlignmentProj()

        correctionFactor = Xdim / self.newXdim

        inverseTransform = partSet.getAlignment() == ALIGN_PROJ

        for idx, particle in enumerate(inputSet.iterItems()):

            outParticle = ParticleFlex(progName=const.NMA)
            outParticle.copyInfo(particle)

            outParticle.setZFlex(nma_space[idx])

            if refinePose:
                tr_ori = particle.getTransform().getMatrix()
                shifts, angles = geometryFromMatrix(tr_ori, inverseTransform)

                # Apply delta angles
                angles[0] += delta_rot[idx]
                angles[1] += delta_tilt[idx]
                angles[2] += delta_psi[idx]

                # Apply delta shifts
                shifts[0] += correctionFactor * delta_shift_x[idx]
                shifts[1] += correctionFactor * delta_shift_y[idx]

                # Set new transformation matrix
                tr = matrixFromGeometry(shifts, angles, inverseTransform)
                outParticle.getTransform().setMatrix(tr)

            partSet.append(outParticle)

        partSet.getFlexInfo().n_modes = Integer(deepNMAProtocol.n_modes.get())
        partSet.getFlexInfo().modelPath = String(model_path)
        partSet.getFlexInfo().atomSubset = String(self._subset[deepNMAProtocol.atomSubset.get()])

        structure = deepNMAProtocol.inputStruct.get().getFileName()
        partSet.getFlexInfo().refStruct = String(structure)

        partSet.getFlexInfo().refPose = refinePose

        self._defineOutputs(outputParticles=partSet)
        self._defineTransformRelation(self.inputParticles, partSet)

    # --------------------------- UTILS functions -----------------------
    def _updateParticle(self, item, row):
        setXmippAttributes(item, row, md.MDL_ANGLE_ROT, md.MDL_ANGLE_TILT,
                           md.MDL_ANGLE_PSI, md.MDL_SHIFT_X, md.MDL_SHIFT_Y,
                           md.MDL_FLIP)
        createItemMatrix(item, row, align=ALIGN_PROJ)

    def getInputParticles(self):
        return self.inputParticles.get()

    def readPDB(self, fnIn):
        with open(fnIn) as f:
            lines = f.readlines()
        return lines

    def PDB2List(self, lines):
        deepNMAProtocol = self.deepNMAProtocol.get()
        newlines = []
        for line in lines:
            eval = re.search(r'^ATOM\s+\d+\s+/N|CA|C|O/\s+', line) \
                if deepNMAProtocol.onlyBackbone.get() else line.startswith("ATOM ")
            if eval:
                try:
                    x = float(line[30:38])
                    y = float(line[38:46])
                    z = float(line[46:54])
                    val = float(line[54:60])
                    newline = [x, y, z, val]
                    newlines.append(newline)
                except:
                    pass
        return newlines

    # ----------------------- VALIDATE functions ----------------------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        return errors



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os

import numpy as np
import pandas as pd

from pyworkflow import NEW
from pyworkflow.protocol.params import PointerParam, IntParam, EnumParam

from pwem.protocols import ProtAnalysis3D

import flexutils


class ProtFlexOptimalClusters(ProtAnalysis3D):
    """ Optimal cluster number analysis for flexibility spaces """

    _label = 'find optimal clusters'
    _devStatus = NEW
    OUTPUT_PREFIX = 'selectedReference'
    CHOICES = ["KMeans"]

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='General parameters')
        form.addParam('inputParticles', PointerParam, label="Input particles",
                      pointerClass='SetOfParticlesFlex', important=True,
                      help="Particles with the flexibility space to be analyzed")
        form.addParam('maxClusters', IntParam, default=15, label="Maximum number of clusters",
                      help="Determines up to how many clusters the analysis will be carried "
                           "out")
        form.addParam('clusterMethod', EnumParam, default=0, choices=self.CHOICES,
                      label="Clustering method", display=EnumParam.DISPLAY_COMBO,
                      help="Determines the clustering method that will be used to analyze "
                           "the flexible space")
        form.addParallelSection(threads=4, mpi=0)

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.clusterAnalysis)

    # --------------------------- STEPS functions -----------------------------
    def clusterAnalysis(self):
        inputParticles = self.inputParticles.get()
        maxClusters = self.maxClusters.get()
        clusterMethod = self.CHOICES[self.clusterMethod.get()]
        outPath = self._getExtraPath()
        dataPath = self._getExtraPath("flex_space.txt")

        # Save flex space to file
        flex_space = []
        for particle in inputParticles.iterItems():
            flex_space.append(particle.getZFlex())
        flex_space = np.asarray(flex_space)
        np.savetxt(dataPath, flex_space)

        # Optimal cluster analysis
        args = "--data_file %s --out_path %s --max_clusters %d --cluster_method %s" \
               % (dataPath, outPath, maxClusters, clusterMethod)
        program = flexutils.Plugin.getTensorflowProgram("find_optimal_clusters.py", python=False)
        self.runJob(program, args, numberOfMpi=1)


    # --------------------------- UTILS functions ----------------------------

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        if os.path.isfile(self._getExtraPath("auto_clustering_results.csv")):
            summary.append("Analysis results (optimal cluster number):")
            df = pd.read_csv(self._getExtraPath("auto_clustering_results.csv"))
            for _, row in df.iterrows():
                summary.append(f"     - {row['Method']}: {row['Best_K']}")
        else:
            summary.append("Finding optimal number of clusters...")
        return summary

    def _methods(self):
        return [
            "Optimal cluster analysis based on the following methods: \n"
            "     - Gap statistic (gapStatistic)\n"
            "     - Elbow (elbow)\n"
            "     - Silhouette (silhouette)\n"
            "     - Calinski Harabasz (ch)\n"
            "     - Davies Bouldin (db)\n"
        ]



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os, matplotlib, math
from scipy import ndimage
from scipy.spatial import KDTree
import tkinter as tk
matplotlib.use("TkAgg")
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.widgets import RadioButtons
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from mpl_toolkits.mplot3d.proj3d import proj_transform
from matplotlib.text import Annotation

import numpy as np
from scipy.ndimage.filters import gaussian_filter

from pyworkflow.viewer import DESKTOP_TKINTER, WEB_DJANGO, ProtocolViewer
from pyworkflow.gui.plotter import plt
import pyworkflow.protocol.params as params

from flexutils.protocols.xmipp.protocol_structure_map_zernike3d import XmippProtStructureMapZernike3D


class XmippProtStructureMapZernike3DViewer(ProtocolViewer):
    """ Wrapper to visualize different type of data objects
    with the Xmipp program xmipp_showj
    """
    
    _label = 'viewer structure map'
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]
    _targets = [XmippProtStructureMapZernike3D]

    def _defineParams(self, form):
        form.addSection(label='Show StructMap')
        form.addParam('map', params.EnumParam, choices=['Deformation', 'Correlation', 'Consensus'],
                      default=0, help='Choose the type of metric to display the coordinates.')
        form.addParam('twoSets', params.BooleanParam, default=False, label='Two set analysis',
                      help='Activate if the analysis of the deformation included the option of analysing '
                           'two sets of volumes independently.')
        form.addParam('doShowPlot', params.LabelParam,
                      label="Display the StructMap")
    
    def _getVisualizeDict(self):
        return {'doShowPlot': self._visualize}

    def getOutputFile(self):
        fnOutput = ['']
        if self.map.get() == 0 and not self.twoSets:
            fnOutput = [self.protocol._defineResultsName(3)]
        elif self.map.get() == 1 and not self.twoSets:
            fnOutput = [self.protocol._defineResultsName2(3)]
        elif self.map.get() == 0 and self.twoSets:
            fnOutput = [self.protocol._defineResultsName(3, 'Sub_1_'),
                        self.protocol._defineResultsName(3, 'Sub_2_')]
        elif self.map.get() == 1 and self.twoSets:
            fnOutput = [self.protocol._defineResultsName2(3, 'Sub_1_'),
                        self.protocol._defineResultsName2(3, 'Sub_2_')]
        elif self.map.get() == 2 and not self.twoSets:
            fnOutput = [self.protocol._defineResultsName3(3)]
        return fnOutput

    def _visualize(self, e=None):
        fnOutput = self.getOutputFile()
        for file in fnOutput:
            if not os.path.exists(file):
                return [self.errorMessage('The necessary metadata was not produced\n'
                                          'Execute again the protocol\n',
                                          title='Missing result file')]

        self.coordinates = (np.loadtxt(file) for file in fnOutput)
        self.coordinates = np.vstack(self.coordinates)
        labels = [str(idp) for idp in range(1, self.coordinates.shape[0] + 1)]
        if os.path.isfile(self._getExtraPath('weigths.txt')):
            weights = np.loadtxt(self._getExtraPath('weigths.txt'))
        else:
            weights = None
        plot = projectionPlot(self.coordinates, labels, weights)
        plot.initializePlot()
        return plot
        
    def _validate(self):
        errors = []
        return errors


class projectionPlot(object):

    def __init__(self, coords, labels, weights):
        self.coords = coords
        self.labels = labels
        self.weights = weights
        try:
            self.minimum_spanning_tree()
        except IndexError:
            self.T = None
        self.proj_coords = None
        self.radio = None
        self.cb = None
        self.prevlabel = 'Scatter'
        self.root = tk.Tk()
        self.fig = plt.Figure(figsize=(10, 4))
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.root)
        self.ax_3d = self.fig.add_subplot(121, projection="3d")
        self.ax_3d.set_title("3D view")
        self.ax_3d.set_axis_off()
        self.fig.canvas.mpl_connect('button_release_event', self.onRelease)
        self.ax_2d = self.fig.add_subplot(122)
        self.ax_2d.set_title("Projection from current 3D view")

    def minimum_spanning_tree(self):
        N = self.coords.shape[0]
        tree = KDTree(self.coords)
        distances, indices = tree.query(self.coords, k=10)
        nn_mat = np.zeros((N, N))
        for idn in range(N):
            nn_mat[idn, indices[idn]] += distances[idn].reshape(-1)
            nn_mat[idn, idn] = 0

        edges = ((int(e[0]), int(e[1])) for e in zip(*np.asarray(nn_mat).nonzero()))
        triples = list(((u, v, float(nn_mat[u, v])) for u, v in edges))
        edges = [(triple[0], triple[1]) for triple in triples]
        weigths_edge = [triple[2] for triple in triples]

        edge_matrix = np.zeros((N, N))
        for edge, weigth in zip(edges, weigths_edge):
            edge_matrix[edge] = weigth
        self.T = self.KruskalMST(triples, N)

    def KruskalMST(self, triples, N):
        result = []
        i = 0
        e = 0
        graph = sorted(triples, key=lambda item: item[2])
        parent = []
        rank = []
        for node in range(N):
            parent.append(node)
            rank.append(0)
        while e < N - 1:
            u, v, w = graph[i]
            i = i + 1
            x = self.find(parent, u)
            y = self.find(parent, v)
            if x != y:
                e = e + 1
                result.append([u, v, w])
                self.union(parent, rank, x, y)
        return result

    def find(self, parent, i):
        if parent[i] == i:
            return i
        return self.find(parent, parent[i])

    def union(self, parent, rank, x, y):
        xroot = self.find(parent, x)
        yroot = self.find(parent, y)
        if rank[xroot] < rank[yroot]:
            parent[xroot] = yroot
        elif rank[xroot] > rank[yroot]:
            parent[yroot] = xroot
        else:
            parent[yroot] = xroot
            rank[xroot] += 1

    def mst_3D(self):
        N = self.coords.shape[0]
        degree = [0] * N
        edges_mst = [0] * (N - 1)
        if self.T is not None:
            for idn in range(N - 1):
                degree[self.T[idn][0]] += 1
                degree[self.T[idn][1]] += 1
                edges_mst[idn] = (self.T[idn][0], self.T[idn][1])
            edge_max = max(degree)
            colors = [plt.cm.plasma(val / edge_max) for val in degree]
            for edge in edges_mst:
                if edge != 0:
                    x = np.array((self.coords[edge[0]][0], self.coords[edge[1]][0]))
                    y = np.array((self.coords[edge[0]][1], self.coords[edge[1]][1]))
                    z = np.array((self.coords[edge[0]][2], self.coords[edge[1]][2]))
                    self.ax_3d.plot(x, y, z, c='black', alpha=0.5)
        else:
            colors = [plt.cm.plasma(1) for _ in self.coords]
        for idn, row in enumerate(self.coords):
            xi = row[0]
            yi = row[1]
            zi = row[2]
            self.ax_3d.scatter(xi, yi, zi, c=[colors[idn]], s=20 + 20 * degree[idn], edgecolors='k', alpha=0.7)
        annotate3D(self.ax_3d, s=self.labels, xyz=self.coords, fontsize=10, xytext=(-3, 3),
                   textcoords='offset points', ha='right', va='bottom')

    def projectMatrix(self, M, coords):
        proj_coords = []
        for coord in coords:
            coord = np.append(coord, 1.0)
            proj_coord = M.dot(coord)
            proj_coords.append(proj_coord)
        return np.asarray(proj_coords)

    def onRelease(self, event):
        if event.inaxes == self.ax_3d:
            M = event.inaxes.get_proj()
            self.proj_coords = self.projectMatrix(M, self.coords)
            self.plotType(self.prevlabel)

    def plotScatter(self, x, y):
        self.ax_2d.clear()
        self.ax_2d.scatter(x, y, color="green")
        self.ax_2d.set_title("Projection Minimum Spanning Tree")
        self.fig.canvas.draw()

    def plotContour(self, x, y):
        self.ax_2d.clear()
        rangeX = np.max(x) - np.min(x)
        rangeY = np.max(y) - np.min(y)
        if rangeX > rangeY:
            sigma = rangeX / 50
        else:
            sigma = rangeY / 50
        xi = np.linspace(min(x) - 0.1, max(x) + 0.1, 100)
        yi = np.linspace(min(x) - 0.1, max(x) + 0.1, 100)
        z = np.zeros((100, 100), float)
        zSize = z.shape
        N = len(x)
        for c in range(zSize[1]):
            for r in range(zSize[0]):
                for d in range(N):
                    z[r, c] = z[r, c] + (1.0 / N) * (1.0 / ((2 * math.pi) * sigma ** 2)) * math.exp(
                        -((xi[c] - x[d]) ** 2 + (yi[r] - y[d]) ** 2) / (2 * sigma ** 2))
        zMax = np.max(z)
        z = z / zMax
        self.ax_2d.contour(xi, yi, z, 15, linewidths=0.5, colors='k')
        cf = self.ax_2d.contourf(xi, yi, z, 15, cmap=plt.cm.jet)
        self.ax_2d.set_title("Projection Scatter Plot")
        cbaxes = self.fig.add_axes([0.92, 0.1, 0.01, 0.8])
        self.cb = self.fig.colorbar(mappable=cf, cax=cbaxes)
        self.cb.set_ticks([])
        self.fig.canvas.draw()

    def plotConvolution(self, x, y):
        coordinates = np.stack((x, y), axis=1)
        Xr = np.round(coordinates, decimals=3)
        size_grid = 2 * np.amax(Xr)
        grid_coords = np.arange(-size_grid, size_grid, 0.001)
        R, C = np.meshgrid(grid_coords, grid_coords, indexing='ij')
        S = np.zeros(R.shape)
        sigma = R.shape[0] / (200 / 5)
        lbox = int(6 * sigma)
        if lbox % 2 == 0:
            lbox += 1
        mid = int((lbox - 1) / 2 + 1)
        kernel = np.zeros((lbox, lbox))
        kernel[mid, mid] = 1
        kernel = gaussian_filter(kernel, sigma=sigma)
        for p in range(Xr.shape[0]):
            indx = np.argmin(np.abs(R[:, 0] - Xr[p, 0]))
            indy = np.argmin(np.abs(C[0, :] - Xr[p, 1]))
            if 'weights' in locals():
                S[indx - mid:indx + mid - 1, indy - mid:indy + mid - 1] += kernel * self.weights
            else:
                S[indx - mid:indx + mid - 1, indy - mid:indy + mid - 1] += kernel
        S = S[~np.all(S == 0, axis=1)]
        S = S[:, ~np.all(S == 0, axis=0)]
        S = ndimage.rotate(S, 90)
        cf = self.ax_2d.imshow(S)
        cbaxes = self.fig.add_axes([0.92, 0.1, 0.01, 0.8])
        self.ax_2d.set_title('Projection Scatter Plot')
        self.cb = self.fig.colorbar(mappable=cf, cax=cbaxes)
        self.cb.set_ticks([])
        self.fig.canvas.draw()

    def plotType(self, label):
        self.prevlabel = label
        x = self.proj_coords[:, 0]
        y = self.proj_coords[:, 1]
        if self.cb != None:
            self.cb.remove()
            self.cb = None
        if label == 'Scatter':
            self.plotScatter(x, y)
        elif label == 'Contour':
            self.plotContour(x, y)
        elif label == 'Convolution':
            self.plotConvolution(x, y)

    def initializePlot(self):
        self.mst_3D()
        M = self.ax_3d.get_proj()
        self.proj_coords = self.projectMatrix(M, self.coords)
        self.ax_2d.scatter(self.proj_coords[:, 0], self.proj_coords[:, 1], color="green")

        # Buttons
        axcolor = 'silver'
        rax = self.fig.add_axes([0.01, 0.4, 0.12, 0.25], facecolor=axcolor)
        self.radio = RadioButtons(rax, ('Scatter', 'Contour', 'Convolution'))
        self.radio.on_clicked(self.plotType)
        self.canvas.get_tk_widget().pack(side=tk.BOTTOM, fill=tk.BOTH, expand=1)
        self.canvas._tkcanvas.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=1)

        # Toolbar
        toolbar = NavigationToolbar2Tk(self.canvas, self.root)
        toolbar.update()
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        tk.mainloop()

class Annotation3D(Annotation):
    '''Annotate the point xyz with text s'''

    def __init__(self, s, xyz, *args, **kwargs):
        Annotation.__init__(self, "", xy=(0, 0), bbox=dict(boxstyle="round,pad=0.3", fc="whitesmoke", ec="lightgray", lw=2),
                            *args, **kwargs)
        self._verts3d = xyz
        self.s = s

    def draw(self, renderer):
        for coord, text in zip(self._verts3d, self.s):
            xs3d, ys3d, zs3d = coord[0], coord[1], coord[2]
            xs, ys, zs = proj_transform(xs3d, ys3d, zs3d, renderer.M)
            self.xy = (xs, ys)
            Annotation.set_text(self, text)
            Annotation.draw(self, renderer)

def annotate3D(ax, s, *args, **kwargs):
    '''add anotation text s to to Axes3d ax'''

    tag = Annotation3D(s, *args, **kwargs)
    ax.add_artist(tag)



# **************************************************************************
# *
# * Authors:  David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os

from pyworkflow.object import Set
from pyworkflow.viewer import DESKTOP_TKINTER, WEB_DJANGO, ProtocolViewer
import pyworkflow.protocol.params as params
from pyworkflow.utils import removeExt

from pwem.viewers import ChimeraView
from pwem import objects

from flexutils.protocols.xmipp.protocol_apply_field_zernike3d import XmippApplyFieldZernike3D


class XmippApplyFieldZernike3DView(ProtocolViewer):
    """ Visualize a Zernike3D map/structure """
    _label = 'viewer apply field Zernike3D'
    # _targets = [XmippProtVolumeDeformZernike3D, XmippApplyFieldZernike3D]
    _targets = [XmippApplyFieldZernike3D]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]
    OPEN_FILE = "open %s\n"
    VOXEL_SIZE = "volume #%d voxelSize %s\n"
    VOL_HIDE = "vol #%d hide\n"
    VIEW = "view\n"

    def __init__(self, **kwargs):
        ProtocolViewer.__init__(self, **kwargs)
        if type(self.protocol).__name__ == XmippApplyFieldZernike3D.__name__:
            self.deformed = self.protocol.deformed
            self.have_set = isinstance(self.deformed, Set)
            if self.have_set:
                self.choices = list(self.deformed.getIdSet())
            else:
                self.choices = [0]
                self.deformed = [self.deformed]
        else:
            self.have_set = False
            self.choices = None

        if isinstance(self.deformed, objects.SetOfVolumesFlex) or isinstance(self.deformed[0], objects.VolumeFlex):
            self.mode = "Map"
        else:
            self.mode = "Structure"

    def _defineParams(self, form):
        form.addSection(label='Show deformation')
        form.addParam('idChoice', params.EnumParam,
                      condition='have_set',
                      choices=[str(idx) for idx in self.choices], default=0,
                      label=f'{self.mode} to display', display=params.EnumParam.DISPLAY_COMBO,
                      help=f'Select which {self.mode} to display from the IDs of the set')
        form.addParam('doShowStrain', params.LabelParam,
                      label="Display the strain deformation")
        form.addParam('doShowRotation', params.LabelParam,
                      label="Display the rotation deformation")
        if self.mode == "Map":
            form.addParam('doShowMorphOrigRef', params.LabelParam,
                          label="Display the morphing between original and reference structure")
            form.addParam('doShowMorphDeformedRef', params.LabelParam,
                          label="Display the morphing between deformed and reference structure")
        elif self.mode == "Structure":
            form.addParam('doShowPDB', params.LabelParam,
                          label="Display original and deformed PDB or volume")
            form.addParam('doShowMorph', params.LabelParam,
                          label="Display a morphing between the original and deformed structure")

    def _getVisualizeDict(self):
        if type(self.protocol).__name__ == XmippApplyFieldZernike3D.__name__:
            if self.have_set:
                self.chosen = self.deformed[self.choices[self.idChoice.get()]]
            else:
                self.chosen = self.deformed[0]

            if self.mode == "Map":
                self.inputVol = self.protocol.inputVolume.get() if not self.have_set else \
                                self.protocol.inputVolume.get()[list(self.deformed.getIdSet())[self.idChoice.get()]]
                myDict = {'fnRefVol': self.inputVol.getFileName(),
                          'fnOutVol': self.chosen.getFileName()}

                self.protocol._updateFilenamesDict(myDict)
                # if not self.have_set:
                #     self.protocol._createFilenameTemplates()

        if self.mode == "Map":
            return {'doShowStrain': self._doShowStrain,
                    'doShowRotation': self._doShowRotation,
                    'doShowMorphOrigRef': self._doShowMorphOrigRef,
                    'doShowMorphDeformedRef': self._doShowDeformedOrigRef}
        elif self.mode == "Structure":
            return {'doShowStrain': self._doShowStrainStruct,
                    'doShowRotation': self._doShowRotationStruct,
                    'doShowPDB': self._doShowPDB,
                    'doShowMorph': self._doShowMorph}

    # ------------------- Mode MAP Methods -------------------
    def _doShowStrain(self, param=None):
        scriptFile = self.protocol._getPath('strain_chimera.cxc')
        fhCmd = open(scriptFile, 'w')
        fnref = os.path.abspath(self.chosen.getFileName())
        smprt = self.chosen.getSamplingRate()

        fnbase2 = removeExt(self.chosen.getFileName())
        fnStrain = os.path.abspath(fnbase2)

        fhCmd.write(self.OPEN_FILE % fnref)
        fhCmd.write(self.OPEN_FILE % (fnStrain + "_strain.mrc"))
        counter = 1
        fhCmd.write(self.VOXEL_SIZE % (counter, str(smprt)))
        counter += 1
        fhCmd.write(self.VOXEL_SIZE % (counter, str(smprt)))
        fhCmd.write(self.VOL_HIDE % counter)
        fhCmd.write('color sample #%d map #%d palette rainbow\n' % (counter - 1, counter))
        fhCmd.write(self.VIEW)
        fhCmd.close()

        view = ChimeraView(scriptFile)
        return [view]

    def _doShowRotation(self, param=None):
        scriptFile = self.protocol._getPath('rotation_chimera.cxc')
        fhCmd = open(scriptFile, 'w')
        fnref = os.path.abspath(self.chosen.getFileName())
        smprt = self.chosen.getSamplingRate()

        fnbase2 = removeExt(self.chosen.getFileName())
        fnStrain = os.path.abspath(fnbase2)

        fhCmd.write(self.OPEN_FILE % fnref)
        fhCmd.write(self.OPEN_FILE % (fnStrain + "_rotation.mrc"))
        counter = 1
        fhCmd.write(self.VOXEL_SIZE % (counter, str(smprt)))
        counter += 1
        fhCmd.write(self.VOXEL_SIZE % (counter, str(smprt)))
        fhCmd.write(self.VOL_HIDE % (counter))
        fhCmd.write('color sample #%d map #%d palette rainbow\n' % (counter - 1, counter))
        fhCmd.write(self.VIEW)
        fhCmd.close()

        view = ChimeraView(scriptFile)
        return [view]

    def _doShowMorphOrigRef(self, param=None):
        scriptFile = self.protocol._getPath('morph_orig_ref_chimera.cxc')
        fhCmd = open(scriptFile, 'w')
        fninput = os.path.abspath(self.inputVol.getFileName())
        fnref = os.path.abspath(self.chosen.getFlexInfo().refMap.get())
        smprt = self.chosen.getSamplingRate()

        fhCmd.write(self.OPEN_FILE % fninput)
        fhCmd.write(self.OPEN_FILE % fnref)

        counter = 1
        fhCmd.write(self.VOXEL_SIZE % (counter, str(smprt)))
        fhCmd.write(self.VOL_HIDE % (counter))
        counter += 1
        fhCmd.write(self.VOXEL_SIZE % (counter, str(smprt)))
        fhCmd.write(self.VOL_HIDE % (counter))
        fhCmd.write("volume morph #%d,%d frames 100 playStep 0.01\n" % (counter - 1, counter))
        fhCmd.write(self.VIEW)
        fhCmd.close()

        view = ChimeraView(scriptFile)
        return [view]

    def _doShowDeformedOrigRef(self, param=None):
        if self.protocol.applyPDB.get():
            raise ValueError("This viewer is only for volumes, not atomic structures")

        scriptFile = self.protocol._getPath('morph_deformed_ref_chimera.cxc')
        fhCmd = open(scriptFile, 'w')
        fninput = os.path.abspath(self.inputVol.getFileName())
        fnref = os.path.abspath(self.chosen.getFileName())
        smprt = self.chosen.getSamplingRate()

        fhCmd.write(self.OPEN_FILE % fninput)
        fhCmd.write(self.OPEN_FILE % fnref)
        counter = 1
        fhCmd.write(self.VOXEL_SIZE % (counter, str(smprt)))
        fhCmd.write(self.VOL_HIDE % counter)
        counter += 1
        fhCmd.write(self.VOXEL_SIZE % (counter, str(smprt)))
        # fhCmd.write("focus\n")
        fhCmd.write(self.VOL_HIDE % counter)
        fhCmd.write("volume morph #%d,%d frames 100 playStep 0.01\n" % (counter-1, counter))
        fhCmd.write(self.VIEW)
        fhCmd.close()

        view = ChimeraView(scriptFile)
        return [view]
    # ------------------- ------------------- -------------------

    # ------------------- Mode Structure Methods -------------------
    def _doShowStrainStruct(self, param=None):
        scriptFile = self.protocol._getPath('strain_chimera.cxc')
        fhCmd = open(scriptFile, 'w')

        fnbase = removeExt(self.chosen.getFileName())
        fnStrain = os.path.abspath(fnbase)

        fhCmd.write(self.OPEN_FILE % (fnStrain + "_strain.pdb"))
        fhCmd.write("show cartoons\n")
        fhCmd.write("cartoon style width 1.5 thick 1.5\n")
        fhCmd.write("style stick\n")
        fhCmd.write('color by occupancy palette rainbow\n')
        fhCmd.write(self.VIEW)
        fhCmd.close()

        view = ChimeraView(scriptFile)
        return [view]

    def _doShowRotationStruct(self, param=None):
        scriptFile = self.protocol._getPath('strain_chimera.cxc')
        fhCmd = open(scriptFile, 'w')

        fnbase = removeExt(self.chosen.getFileName())
        fnStrain = os.path.abspath(fnbase)

        fhCmd.write(self.OPEN_FILE % (fnStrain + "_rotation.pdb"))
        fhCmd.write("show cartoons\n")
        fhCmd.write("cartoon style width 1.5 thick 1.5\n")
        fhCmd.write("style stick\n")
        fhCmd.write('color by occupancy palette rainbow\n')
        fhCmd.write(self.VIEW)
        fhCmd.close()

        view = ChimeraView(scriptFile)
        return [view]

    def _doShowPDB(self, obj, **kwargs):
        if self.protocol.applyPDB.get():
            scriptFile = self.protocol._getPath('pdb_deform_chimera.cxc')
            fhCmd = open(scriptFile, 'w')
            inputFile = os.path.abspath(self.protocol.inputPDB.get().getFileName())
            outputFile = os.path.abspath(self.chosen.getFileName())

            fhCmd.write(self.OPEN_FILE % inputFile)
            fhCmd.write(self.OPEN_FILE % outputFile)
            # fhCmd.write("start Model Panel\n")
            fhCmd.write("show cartoons\n")
            fhCmd.write("cartoon style width 1.5 thick 1.5\n")
            fhCmd.write("style stick\n")
            fhCmd.write("color bymodel\n")
            fhCmd.close()

            view = ChimeraView(scriptFile)
            return [view]
        else:
            raise ValueError("This viewer is only for atomic structures")

    def _doShowMorph(self, obj, **kwargs):
        if self.protocol.applyPDB.get():
            scriptFile = self.protocol._getPath('pdb_deform_chimera.cxc')
            fhCmd = open(scriptFile, 'w')
            inputFile = os.path.abspath(self.protocol.inputPDB.get().getFileName())
            outputFile = os.path.abspath(self.chosen.getFileName())

            fhCmd.write(self.OPEN_FILE % inputFile)
            fhCmd.write(self.OPEN_FILE % outputFile)
            fhCmd.write("hide models\n")
            fhCmd.write("morph #1,2 frames 50 play false\n")
            fhCmd.write("coordset #3 1,\n")
            fhCmd.write("wait 50\n")
            fhCmd.write("coordset #3 50,1\n")
            fhCmd.close()

            view = ChimeraView(scriptFile)
            return [view]
        else:
            raise ValueError("This viewer is only for atomic structures")
    # ------------------- ------------------- -------------------

